# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from . import api



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from ezfastq.copier import FastqCopier
from importlib.resources import files
import pytest

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


SEQ_PATH_1 = files("ezfastq") / "tests" / "data" / "flat"
SEQ_PATH_2 = files("ezfastq") / "tests" / "data" / "nested"


def test_copier_basic():
    sample_names = ["test1", "test2"]
    copier = FastqCopier.from_dir(sample_names, SEQ_PATH_1)
    observed = [fqfile.source_path.name for fqfile in copier]
    expected = [
        "test1_S1_L001_R1_001.fastq.gz",
        "test1_S1_L001_R2_001.fastq.gz",
        "test2_R1.fq.gz",
        "test2_R2.fq.gz",
    ]
    print(observed)
    assert observed == expected


def test_copier_copy(tmp_path):
    sample_names = ["test1", "test2"]
    # First pass: copy all 4
    copier1 = FastqCopier.from_dir(sample_names, SEQ_PATH_1)
    copier1.copy_files(tmp_path)
    assert len(copier1.copied_files) == 4
    assert len(copier1.skipped_files) == 0
    assert len(list(tmp_path.glob("*_R?.fastq.gz"))) == 4
    # Second pass: none are re-copied
    copier2 = FastqCopier.from_dir(sample_names, SEQ_PATH_1)
    copier2.copy_files(tmp_path)
    assert len(copier2.copied_files) == 0
    assert len(copier2.skipped_files) == 4
    # Third pass: delete one and make sure it is the only one re-copied
    (tmp_path / "test2_R2.fastq.gz").unlink()
    copier3 = FastqCopier.from_dir(sample_names, SEQ_PATH_1)
    copier3.copy_files(tmp_path)
    assert len(copier3.copied_files) == 1
    assert copier3.copied_files[0].name == "test2_R2.fastq.gz"
    assert len(copier3.skipped_files) == 3


def test_copier_prefix(tmp_path):
    sample_names = ["test2", "test3"]
    copier = FastqCopier.from_dir(sample_names, SEQ_PATH_1, prefix="abc_")
    copier.copy_files(tmp_path)
    assert len(list(tmp_path.glob("abc_*.fastq.gz"))) == 4


def test_copier_str_basic(tmp_path):
    sample_names = ["test1", "test2", "test3"]
    copier = FastqCopier.from_dir(sample_names, SEQ_PATH_1)
    copier.copy_files(tmp_path)
    observed = str(copier)
    expected = """
[CopiedFiles]
"test1_S1_L001_R1_001.fastq.gz" = "test1_R1.fastq.gz"
"test1_S1_L001_R2_001.fastq.gz" = "test1_R2.fastq.gz"
"test2_R1.fq.gz" = "test2_R1.fastq.gz"
"test2_R2.fq.gz" = "test2_R2.fastq.gz"
"test3-reads-r1.fastq" = "test3_R1.fastq.gz"
"test3-reads-r2.fastq" = "test3_R2.fastq.gz"
"""
    assert observed.strip() == expected.strip()


def test_copier_str_noop(tmp_path):
    sample_names = ["test1", "test2", "test3"]
    copier = FastqCopier.from_dir(sample_names, SEQ_PATH_1)
    assert str(copier) == ""


def test_copier_str_allskip(tmp_path):
    sample_names = ["test1"]
    copier = FastqCopier.from_dir(sample_names, SEQ_PATH_1)
    (tmp_path / "test1_R1.fastq.gz").touch()
    (tmp_path / "test1_R2.fastq.gz").touch()
    copier.copy_files(tmp_path)
    observed = str(copier)
    expected = """
[SkippedFiles]
already_copied = [
    "test1_S1_L001_R1_001.fastq.gz",
    "test1_S1_L001_R2_001.fastq.gz",
]
"""
    assert observed.strip() == expected.strip()


def test_copier_str_mixed(tmp_path):
    sample_names = ["test1", "test2", "test3"]
    copier = FastqCopier.from_dir(sample_names, SEQ_PATH_2)
    (tmp_path / "test2_R1.fastq.gz").touch()
    (tmp_path / "test2_R2.fastq.gz").touch()
    copier.copy_files(tmp_path)
    observed = str(copier)
    expected = """
[CopiedFiles]
"test1_S1_L001_R1_001.fastq.gz" = "test1_R1.fastq.gz"
"test1_S1_L001_R2_001.fastq.gz" = "test1_R2.fastq.gz"
"test3-reads-r1.fastq" = "test3_R1.fastq.gz"
"test3-reads-r2.fastq" = "test3_R2.fastq.gz"

[SkippedFiles]
already_copied = [
    "test2_R1.fq.gz",
    "test2_R2.fq.gz",
]
"""
    assert observed.strip() == expected.strip()


def test_copier_str_roundtrip(tmp_path):
    sample_names = ["test1", "test2", "test3"]
    copier = FastqCopier.from_dir(sample_names, SEQ_PATH_1)
    (tmp_path / "test2_R1.fastq.gz").touch()
    (tmp_path / "test2_R2.fastq.gz").touch()
    copier.copy_files(tmp_path)
    copy_log = tmp_path / "log.toml"
    with open(copy_log, "w") as fh:
        print(copier, file=fh)
    with open(copy_log, "rb") as fh:
        copy_data = tomllib.load(fh)
    assert len(copy_data["CopiedFiles"]) == 4
    assert copy_data["CopiedFiles"]["test1_S1_L001_R1_001.fastq.gz"] == "test1_R1.fastq.gz"
    observed = copy_data["SkippedFiles"]["already_copied"]
    expected = ["test2_R1.fq.gz", "test2_R2.fq.gz"]
    assert observed == expected



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from .pair import PairMode
from .scanner import FastqFileScanner
from collections import defaultdict


class SampleFastqMap(defaultdict):
    "Map sample names to lists of corresponding FASTQ files."

    @classmethod
    def new(cls, sample_names, data_path, pair_mode=PairMode.Unspecified):
        files_by_sample = cls(list)
        scanner = FastqFileScanner.new(sample_names)
        for sample_name, fastq in scanner.scan(data_path):
            files_by_sample[sample_name].append(fastq)
            files_by_sample[sample_name].sort()
        cls.validate_sample_files(sample_names, files_by_sample, pair_mode=pair_mode)
        return files_by_sample

    @staticmethod
    def validate_sample_files(sample_names, files_by_sample, pair_mode=PairMode.Unspecified):
        for sample in sample_names:
            fastq_files = files_by_sample[sample]
            pair_mode.check(len(fastq_files), sample)



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from ezfastq.map import SampleFastqMap
from ezfastq.pair import PairMode
from importlib.resources import files
import pytest


def test_seqmap_basic(simple_case):
    seq_map = SampleFastqMap.new(["1-1", "1-2", "1-3", "1-4"], simple_case)
    expected = [
        simple_case / "1-1_S1_L001_R1_001.fastq.gz",
        simple_case / "1-1_S1_L001_R2_001.fastq.gz",
    ]
    observed = seq_map["1-1"]
    assert observed == expected
    assert "1-2" in seq_map
    assert "1-3" in seq_map
    assert "1-4" in seq_map
    assert "1-5" not in seq_map
    assert len(list(seq_map)) == 4


def test_seqmap_single(single_end_case):
    seq_map = SampleFastqMap.new(
        ["1-1", "1-2", "1-3"], single_end_case, pair_mode=PairMode.SingleEnd
    )
    assert seq_map["1-1"] == [single_end_case / "1-1_S1_L001_R1_001.fastq.gz"]
    assert "1-2" in seq_map
    assert "1-3" in seq_map
    assert "1-4" not in seq_map


def test_sample_not_found(simple_case):
    message = "sample 1-5: found 0 FASTQ files"
    with pytest.raises(FileNotFoundError, match=message):
        SampleFastqMap.new(["1-1", "1-2", "1-3", "1-4", "1-5"], simple_case)


def test_single_end_mismatch(single_end_case):
    message = r"sample 1-1: found 1 FASTQ file\(s\), expected 2 in paired-end mode"
    with pytest.raises(ValueError, match=message):
        SampleFastqMap.new(["1-1", "1-2", "1-3"], single_end_case, pair_mode=PairMode.PairedEnd)


def test_nested_sample_names():
    sample_names = [f"sample{n+1}" for n in range(12)]
    message = "one sample name cannot be a substring of another sample name: sample1 vs sample10"
    with pytest.raises(ValueError, match=message):
        SampleFastqMap.new(sample_names, ".")


def test_missing_dir():
    sample_names = [f"sample{n+1}" for n in range(9)]
    with pytest.raises(FileNotFoundError):
        SampleFastqMap.new(sample_names, "/a/b/c/BogusDir")


def test_dir_is_file():
    sample_names = [f"sample{n+1}" for n in range(9)]
    file_path = files("ezfastq") / "tests" / "data" / "flat" / "test2_R1.fq.gz"
    with pytest.raises(NotADirectoryError):
        SampleFastqMap.new(sample_names, file_path)


@pytest.fixture
def simple_case(tmp_path):
    for sample in range(1, 5):
        for end in (1, 2):
            fastq = tmp_path / f"1-{sample}_S{sample}_L001_R{end}_001.fastq.gz"
            fastq.touch()
    return tmp_path


@pytest.fixture
def single_end_case(tmp_path):
    for sample in range(1, 4):
        fastq = tmp_path / f"1-{sample}_S{sample}_L001_R1_001.fastq.gz"
        fastq.touch()
    return tmp_path



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from .api import copy
from .pair import PairMode
from argparse import ArgumentParser
from importlib.metadata import version
from pathlib import Path


def main(arglist=None):
    args = parse_args(arglist)
    copy(
        args.samples,
        args.seq_path,
        pair_mode=args.pair_mode,
        prefix=args.prefix,
        workdir=args.workdir,
    )


def parse_args(arglist=None):
    if arglist:
        arglist = map(str, arglist)
    args = get_parser().parse_args(arglist)
    samples_file = Path(args.samples[0])
    samples_file_exists = samples_file.is_file() or samples_file.is_fifo()
    if len(args.samples) == 1 and samples_file_exists:
        args.samples = samples_file.read_text().strip().split("\n")
    args.pair_mode = PairMode(args.pair_mode)
    return args


def get_parser():
    parser = ArgumentParser(description="Copy FASTQ files and update sample names")
    parser.add_argument(
        "seq_path",
        help="path to directory containing sequences in FASTQ format; subdirectories will be searched recursively",
    )
    parser.add_argument(
        "samples",
        nargs="+",
        help="name of one or more samples to process; can be provided as command-line arguments or as a file with one sample name per line",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"ezfastq v{version('ezfastq')}",
    )
    parser.add_argument(
        "-w",
        "--workdir",
        metavar="WD",
        type=Path,
        default=Path("."),
        help="directory to which input files will be copied and renamed",
    )
    parser.add_argument(
        "-p",
        "--prefix",
        metavar="P",
        default="",
        help="prefix to prepend to sample names; resulting file path will be '{workdir}/seq/{prefix}_{sample}_R{1,2}.fastq.gz'",
    )
    parser.add_argument(
        "-m",
        "--pair-mode",
        metavar="M",
        type=int,
        choices=[0, 1, 2],
        default=0,
        help="specify 1 to indicate that all samples are single-end, or 2 to indicate that all samples are paired-end; by default, read layout is inferred automatically on a per-sample basis",
    )
    return parser



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from dataclasses import dataclass
from itertools import combinations
from pathlib import Path
from typing import List


@dataclass
class FastqFileScanner:
    "Supports recursively scanning a directory for FASTQ files containing the given sample names."

    sample_names: List

    @classmethod
    def new(cls, sample_names):
        sample_names = cls.check_sample_names(sample_names)
        scanner = cls(sample_names)
        return scanner

    @staticmethod
    def check_sample_names(sample_names):
        sample_names = sorted(sample_names)
        for sample1, sample2 in combinations(sample_names, 2):
            if sample1 in sample2 or sample2 in sample1:
                message = f"one sample name cannot be a substring of another sample name: {sample1} vs {sample2}"
                raise ValueError(message)
        return sample_names

    def scan(self, path):
        self.check_scan_path(path)
        valid_suffixes = (".fastq", ".fastq.gz", ".fq", ".fq.gz")
        for file_path in self.traverse(path):
            if not file_path.name.endswith(valid_suffixes):
                continue
            for sample in self.sample_names:
                if sample in file_path.name:
                    yield sample, file_path

    @staticmethod
    def check_scan_path(path):
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(path)
        if not path.is_dir():
            raise NotADirectoryError(path)

    @staticmethod
    def traverse(path):
        path = Path(path)
        if not path.is_dir():  # pragma: no cover
            return
        for file_path in path.rglob("*"):
            if file_path.is_file():
                yield file_path



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from ezfastq import cli
from importlib.resources import files
import pytest
from subprocess import run

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


def test_copy(tmp_path):
    seq_path = files("ezfastq") / "tests" / "data" / "flat"
    arglist = [seq_path, "test1", "test2", "test3", "--workdir", tmp_path]
    cli.main(arglist)
    assert len(list((tmp_path / "seq").glob("*_R?.fastq.gz"))) == 6
    copy_log = tmp_path / "seq" / "copy-log-1.toml"
    with open(copy_log, "rb") as fh:
        log_data = tomllib.load(fh)
    assert len(log_data["CopiedFiles"]) == 6
    assert "SkippedFiles" not in log_data


def test_copy_sample_names_file(tmp_path):
    sample_names_file = tmp_path / "sample-names.txt"
    sample_names_file.write_text("test1\ntest3\ntest2\n")
    seq_path = files("ezfastq") / "tests" / "data" / "nested"
    arglist = [seq_path, sample_names_file, "--workdir", tmp_path]
    cli.main(arglist)
    assert len(list((tmp_path / "seq").glob("*_R?.fastq.gz"))) == 6
    copy_log = tmp_path / "seq" / "copy-log-1.toml"
    with open(copy_log, "rb") as fh:
        log_data = tomllib.load(fh)
    assert len(log_data["CopiedFiles"]) == 6
    assert "SkippedFiles" not in log_data


def test_fq_command(tmp_path):
    seq_path = files("ezfastq") / "tests" / "data" / "nested"
    arglist = ["ezfastq", seq_path, "test1", "test2", "test3", "--workdir", tmp_path]
    run(arglist)
    assert len(list((tmp_path / "seq").glob("*_R?.fastq.gz"))) == 6



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from .fastq import FastqFile
from .map import SampleFastqMap
from .pair import PairMode
from dataclasses import dataclass
from io import StringIO
from pathlib import Path
import rich
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.syntax import Syntax
import sys
from typing import List


@dataclass
class FastqCopier:
    """Recursively search a directory for FASTQ files to copy to another.

    FASTQ file names are streamlined in the process, and read pairing status is validated.
    """

    sample_names: List
    copied_files: List
    skipped_files: List
    file_map: SampleFastqMap
    prefix: str = ""

    @classmethod
    def from_dir(cls, sample_names, data_path, prefix="", pair_mode=PairMode.Unspecified):
        copied_files = list()
        skipped_files = list()
        file_map = SampleFastqMap.new(sample_names, data_path, pair_mode=pair_mode)
        copier = cls(sorted(sample_names), copied_files, skipped_files, file_map, prefix)
        return copier

    def copy_files(self, destination):
        progress = self.get_progress_tracker()
        progress.console.log(f"Copying {len(self)} FASTQ files")
        with progress:
            task = progress.add_task("[bold red]Copying...", total=len(self))
            for fastq in self:
                # There's an exact, closed-form solution to the problem addressed in the next few
                # lines that would involve str.format. I also think it would be much more difficult
                # to interpret that code's intent. -- DSS, 2025-05-27
                llsn = self.length_longest_sample_name
                if llsn < 8:
                    desc = f"[bold red]{fastq.sample:>8s} R{fastq.read}"
                elif llsn < 12:
                    desc = f"[bold red]{fastq.sample:>12s} R{fastq.read}"
                else:
                    desc = f"[bold red]{fastq.sample:>16s} R{fastq.read}"
                progress.update(task, description=desc)
                was_copied = fastq.check_and_copy(destination)
                progress.update(task, advance=1)
                if was_copied:
                    self.copied_files.append(fastq)
                else:
                    self.skipped_files.append(fastq)
            progress.update(task, description="[bold red]Finished!")

    @staticmethod
    def get_progress_tracker():
        return Progress(
            SpinnerColumn(),
            TextColumn("[bold red]{task.description}", justify="right"),
            BarColumn(complete_style="green"),
            TextColumn("[bold red]{task.percentage:>5.1f}%", justify="right"),
            TextColumn("[yellow]Time elapsed:"),
            TimeElapsedColumn(),
            TextColumn("[cyan]Est. time remaining:"),
            TimeRemainingColumn(compact=True),
            refresh_per_second=1,
        )

    def print_copy_log(self, outstream=sys.stderr):
        syntax = Syntax(str(self), "toml", theme="solarized-dark")
        panel = Panel(syntax, expand=False, title="FASTQ Copy Log", title_align="right")
        rich.print(panel, file=outstream)

    @property
    def length_longest_sample_name(self):
        return max(len(sample) for sample in self.sample_names)

    def __len__(self):
        return sum(len(fqfiles) for fqfiles in self.file_map.values())

    def __iter__(self):
        for sample_name, fqfiles in sorted(self.file_map.items()):
            for n, fqfile in enumerate(fqfiles, 1):
                source_path = Path(fqfile).absolute()
                read = 0 if len(fqfiles) == 1 else n
                yield FastqFile(source_path, sample_name, read, self.prefix)

    def __str__(self):
        output = StringIO()
        if len(self.copied_files) > 0:
            print("[CopiedFiles]", file=output)
            for fastq in self.copied_files:
                print(fastq, file=output)
        if len(self.skipped_files) > 0:
            print("\n[SkippedFiles]\nalready_copied = [", file=output)
            for fastq in self.skipped_files:
                print(f'    "{fastq.source_path.name}",', file=output)
            print("]", file=output)
        return output.getvalue().strip()



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from .copier import FastqCopier
from .pair import PairMode
from pathlib import Path


def copy(sample_names, seq_path, pair_mode=PairMode.Unspecified, prefix="", workdir=Path(".")):
    copier = FastqCopier.from_dir(sample_names, seq_path, prefix=prefix, pair_mode=pair_mode)
    copier.copy_files(workdir / "seq")
    copier.print_copy_log()
    nlogs = len(list((workdir / "seq").glob("copy-log-*.toml")))
    with open(workdir / "seq" / f"copy-log-{nlogs + 1}.toml", "w") as fh:
        print(copier, file=fh)
    added_samples = set(fastq.sample for fastq in copier.copied_files)
    added_samples = sorted(added_samples)
    with open(workdir / "samples.txt", "a") as fh:
        print(*added_samples, sep="\n", file=fh)
    return copier



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from dataclasses import dataclass
from pathlib import Path
from shutil import copy
from subprocess import run
from typing import Literal


@dataclass
class FastqFile:
    "Facilitate copying FASTQ files and standardizing FASTQ names internally."

    source_path: Path
    sample: str
    read: Literal[0, 1, 2]  # undefined, R1, R2
    prefix: str = ""

    def __str__(self):
        return f'"{self.source_path.name}" = "{self.name}"'

    def check_and_copy(self, destination):
        destination = Path(destination)
        compressed_copy = destination / self.name
        if compressed_copy.is_file():
            return False
        else:
            self.copy(destination)
            return True

    def copy(self, destination):
        destination.mkdir(parents=True, exist_ok=True)
        file_copy = destination / self._working_name
        copy(self.source_path, file_copy)
        if self.extension == "fastq":
            run(["gzip", str(file_copy)])

    @property
    def name(self):
        return f"{self.stem}.fastq.gz"

    @property
    def stem(self):
        read_designator = f"_R{self.read}" if self.read in (1, 2) else ""
        return f"{self.prefix}{self.sample}{read_designator}"

    @property
    def extension(self):
        return "fastq.gz" if self.source_path.name.endswith("gz") else "fastq"

    @property
    def _working_name(self):
        return f"{self.stem}.{self.extension}"



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from enum import Enum


class PairMode(Enum):
    Unspecified = 0
    SingleEnd = 1
    PairedEnd = 2

    def check(self, num_files, sample):
        if num_files == 0:
            raise FileNotFoundError(f"sample {sample}: found 0 FASTQ files")
        if num_files not in self.expected_num_files:
            exp_file_str = " or ".join(map(str, self.expected_num_files))
            message = f"sample {sample}: found {num_files} FASTQ file(s), expected {exp_file_str} in {self.mode} mode"
            raise ValueError(message)

    @property
    def mode(self):
        if self == PairMode.Unspecified:
            return "unspecified"
        elif self == PairMode.SingleEnd:
            return "single-end"
        else:
            return "paired-end"

    @property
    def expected_num_files(self):
        if self == PairMode.Unspecified:
            return [1, 2]
        elif self == PairMode.SingleEnd:
            return [1]
        else:
            return [2]



# -------------------------------------------------------------------------------------------------
# Copyright (c) 2025, DHS. This file is part of ezfastq: https://github.com/bioforensics/ezfastq.
#
# This software was prepared for the Department of Homeland Security (DHS) by the Battelle National
# Biodefense Institute, LLC (BNBI) as part of contract HSHQDC-15-C-00064 to manage and operate the
# National Biodefense Analysis and Countermeasures Center (NBACC), a Federally Funded Research and
# Development Center.
# -------------------------------------------------------------------------------------------------

from ezfastq.fastq import FastqFile
from importlib.resources import files
from pathlib import Path
import pytest


def test_fastq_file_basic():
    infile = FastqFile(Path("/opt/data/seq/s1-R1.fq"), "sample1", 1)
    assert infile.sample == "sample1"
    assert infile.read == 1
    assert infile.name == "sample1_R1.fastq.gz"
    assert infile.extension == "fastq"
    assert str(infile) == '"s1-R1.fq" = "sample1_R1.fastq.gz"'


@pytest.mark.parametrize(
    "prefix, file_name",
    [
        ("caseXYZ_", "caseXYZ_sample2_R1.fastq.gz"),
        ("ProjXYZ_", "ProjXYZ_sample2_R1.fastq.gz"),
    ],
)
def test_fastq_file_prefix(prefix, file_name):
    infile = FastqFile(Path("/data/runs/XYZ/s2-R1.fq.gz"), "sample2", 1, prefix)
    assert infile.name == file_name
    assert infile.extension == "fastq.gz"


def test_fastq_file_copy(tmp_path):
    inpath = files("ezfastq") / "tests" / "data" / "flat" / "test1_S1_L001_R1_001.fastq.gz"
    infile = FastqFile(inpath, "test1", 1)
    infile.copy(tmp_path)
    file_copy = tmp_path / "test1_R1.fastq.gz"
    assert file_copy.is_file()


def test_fastq_file_check_and_copy(tmp_path):
    destination = tmp_path / "seq"
    inpath = files("ezfastq") / "tests" / "data" / "flat" / "test1_S1_L001_R2_001.fastq.gz"
    infile = FastqFile(inpath, "test1", 2)
    was_copied = infile.check_and_copy(destination)
    assert was_copied
    file_copy = tmp_path / "seq" / "test1_R2.fastq.gz"
    assert file_copy.is_file()
    was_copied = infile.check_and_copy(destination)
    assert not was_copied
